# AUTOGENERATED FILE! PLEASE DON'T EDIT
"""These are pretty standard imports that I always
do. I put it here so that I can do::

    from k1lib.imports import *

and everything will be in place"""
try:
    import torch; from torch import nn, optim
    import torch.nn.functional as F, torch.utils.data as data
except: pass
import matplotlib.pyplot as plt, matplotlib as mpl
import numpy as np, dill as pickle, multiprocessing as mp, concurrent.futures as futures
import math, os, time, sys, random, logging, traceback, re, typing, glob, warnings, asyncio
import dill, json, inspect, xml, base64, io
import functools; from functools import partial, lru_cache
import contextlib; from contextlib import contextmanager
from collections import deque, defaultdict
from typing import List, Tuple, Callable, Union, Iterator, Set, Dict, Any
import k1lib; from k1lib import schedule, graphEqn, mo, knn, fmt, selector,\
viz, Cbs, settings, cli, _k1a, serve, p5, k1ui
from k1lib.cli import *; k1 = k1lib; k1a = _k1a
for e in cli._scatteredClis: globals()[e.__name__] = e
if "py_k1lib_in_applyMp" not in os.environ: k1lib.dontWrap()
plt.rcParams['figure.dpi'] = 100
plt.rcParams["animation.html"] = "jshtml"
from math import e, pi; inf = float("inf"); nan = float("nan"); # this section is for constants
h = 6.62607015e-34; hbar = h/(2*pi); Na = 6.0221408e23; kb = 1.380649e-23
c = 299_792_458; qe = 1.60217663e-19; me = 9.1093837e-31; mn = 1.67262192e-27
e0 = 8.85418782e-12; Dal = u = 1.6605390666e-27; R = 8.3145; sb = 5.670374e-8
if settings.startup.or_patch.numpy: cli.init.patchNumpy()
if settings.startup.or_patch.dict: cli.init.patchDict()
try:
    import pandas as pd
    if settings.startup.or_patch.pandas: cli.init.patchPandas()
except ImportError as e: pass
def dummy():
    """Does nothing. Only here so that you can read source code of this file
and see what's up."""
    pass
try: # these are optional imports
    import torchvision as vision
    import torchvision.datasets as tvDs
    import torchvision.transforms as tf
except: pass
try: import IPython
except: pass
try: import PIL
except: pass
try: from scipy.optimize import fsolve; import scipy
except: pass
try: import einops
except: pass
try: from tqdm import tqdm
except: pass
try: import sympy
except: pass
try: import rdkit; from rdkit import Chem
except: pass
try: import requests
except: pass
try: import seaborn as sns
except: pass