# AUTOGENERATED FILE! PLEASE DON'T EDIT
__all__ = ["str_split"]
import k1lib
def str_split(s, delim):
    answer = []
    beginI = 0; inBlock = False; quoteChar = ''
    for i, c in enumerate(s):
        if c == delim and not inBlock:
            answer.append(s[beginI:i])
            beginI = i+1
        elif not inBlock and (c == '"' or c == "'"):
            inBlock = True
            quoteChar = c
        elif inBlock and c == quoteChar:
            inBlock = False
    answer.append(s[beginI:])
    return answer
def str_has_number(s):
    return any([c.isdigit() for c in s])
_g = globals();
"""py stuff"""
py = k1lib.Object()
for e in __all__:
    py[e] = _g[e]
__all__.append("py")
try:
    import k1a
    str_split = k1a.str_split
    str_alpha_numeric = k1a.str_alpha_numeric
    str_kmers = k1a.str_kmers
    str_has_number = k1a.str_has_number
    k1lib._settings.packages.k1a = True
except:
    k1lib._settings.packages.k1a = False